using System;
using System.IO;
using System.Threading;

namespace mp2enc
{
	/// <summary>
	/// GR[hNX
	/// </summary>
	public class Encoder
	{
		private string ifn; /// ̓t@C
		private string ofn; /// o̓t@C
		private Twolame tl; /// TwoLAMEGR[_[

		private ProgressCallback pgcallback; /// ʒuʒm邽߂̃R[obN
		private FinishedCallback ficallback; /// Iʒm邽߂̃R[obN

		private WaveReader wrd;
		private FileStream ofs;
		private BinaryWriter biw;
		Thread  th; /// Xbh
		private bool running; ///  true

		/// <summary>
		/// RXgN^
		/// </summary>
		/// <param name="pgcallbackDelegate">ʒuʒm邽߂̃R[obN</param>
		/// <param name="ficallbackDelegate">Iʒm邽߂̃R[obN</param>
		public Encoder( 
			ProgressCallback pgcallbackDelegate,
			FinishedCallback ficallbackDelegate )
		{
			pgcallback = pgcallbackDelegate;
			ficallback = ficallbackDelegate;
			running = false;
		}

		~Encoder()
		{
			Cleanup();
		}

		/// <summary>
		/// GR[_̏
		/// ͂߂ɕKĂяoĂ
		/// </summary>
		/// <param name="rd">ǂݍރt@C WaveReader</param>
		/// <returns> true</returns>
		public bool Init(WaveReader rd)
		{
			tl = new Twolame();
			if(!tl.init()) return false;

			int ret = 0;
			ret  = tl.set_in_samplerate(rd.SamplesPerSec);
			ret += tl.set_out_samplerate(rd.SamplesPerSec);
			ret += tl.set_num_channels(rd.Channel);

			return ret == 0 ? true : false; 
		}

		/// <summary>
		/// p[^̏
		/// GR[hJnOɕKĂяo
		/// </summary>
		/// <param name="ep">ݒ肷GR[hp[^\̂ւ̎Q</param>
		/// <returns> true</returns>
		public bool InitParams(ref EncProp ep)
		{
			if(tl == null)
			{
				tl = new Twolame();
				if(!tl.init()) return false;
			}

			if(!SetEncProps(ref ep))      return false;
			if(0 != tl.init_params()) return false;

			return true;
		}

		/// <summary>
		/// GR[hJn
		/// [J[Xbh쐬
		/// </summary>
		/// <param name="fin">̓t@Cւ̃pX</param>
		/// <param name="fout">o̓t@Cւ̃pX</param>
		/// <returns></returns>
		public bool Start(string fin, string fout)
		{
			ifn = fin;
			ofn = fout;

			try
			{
				wrd = new WaveReader(ifn);
				ofs = new FileStream(ofn, FileMode.Create);
				biw = new BinaryWriter(ofs);
			}
			catch(IOException)
			{
				Cleanup();
				return false;
			}

			lock(this)
			{
				th = new Thread(new ThreadStart(ThreadProc));
				th.Start();
				running = true;
			}

			return true;
		}

		/// <summary>
		/// GR[h𒆒f
		/// [J[~
		/// </summary>
		public void Abort()
		{
			if(th != null)
			{
				if(th.IsAlive)
				{
					th.Abort();
					th.Join();

					Cleanup();
				}
			}
		}

		/// <summary>
		/// ₢킹
		/// </summary>
		/// <returns>true Ȃ珈</returns>
		public bool IsRunning()
		{
			lock(this)
			{
				return running;
			}
		}
		
		/// <summary>
		/// GR[ĥ߂̃\[XJ
		/// </summary>
		private void Cleanup()
		{
			ifn = null;
			ofn = null;
			running = false;

			if(tl != null)
			{
				tl.close();
				tl = null;
			}
			if(wrd != null)
			{
				wrd.Close();
				wrd = null;
			}
			if(biw != null)
			{
				biw.Close();
				biw = null;
			}
			if(ofs != null)
			{
				ofs.Close();
				ofs = null;
			}
		}

		/// <summary>
		/// [J[Xbh̃vV[W
		/// </summary>
		private void ThreadProc()
		{
			//obt@ TCY͓K
			byte[] ibuf = new byte[32*36*2*2];
			byte[] obuf = new byte[4096];

			try
			{
				int sz_read;
				int sz_out;

				while((sz_read = wrd.Read(ibuf, 0, ibuf.Length)) > 0)
				{
					sz_out = tl.encode_buffer_interleaved(
						ibuf, sz_read/(wrd.Channel*2), obuf, obuf.Length);

					//󋵂ʒm
					if(pgcallback != null)
						pgcallback(sz_read);
					
					biw.Write(obuf, 0, sz_out);
				}
				sz_out = tl.encode_flush(obuf, obuf.Length);
				biw.Write(obuf, 0, sz_out);
			}
			catch(ThreadAbortException)
			{
				//Abortvꂽ
				Cleanup();
				return;
			}

			//f[^ׂď
			Cleanup();
	
			//Iʒm
			if(ficallback != null)
				ficallback(this);
		}

		/// <summary>
		/// ݂̃GR[hp[^𓾂
		/// </summary>
		/// <returns>݂̃GR[hp[^</returns>
		public EncProp GetEncProps()
		{
			EncProp ep = new EncProp();
			if(tl == null) return ep;

			ep.ATH_level		= tl.get_ATH_level();
			ep.Ancillary_bits	= tl.get_num_ancillary_bits();
			ep.Bitrate			= tl.get_bitrate();
			ep.Copyright		= tl.get_copyright();
			ep.CRC				= tl.get_error_protection();
			ep.Emphasis			= tl.get_emphasis();
			ep.Energy_level_ext	= tl.get_energy_levels();
			ep.InSamplerate		= tl.get_in_samplerate();
			ep.MPEG_mode		= tl.get_mode();
			ep.MPEG_version		= tl.get_mpeg_version();
			ep.Num_channels		= tl.get_num_channels();
			ep.Original			= tl.get_original();
			ep.OutSamplerate	= tl.get_out_samplerate();
			ep.Padding			= tl.get_padding();
			ep.Psymodel			= tl.get_psymodel();
			ep.Quick_count		= tl.get_quick_count();
			ep.Quick_mode		= tl.get_quick_mode();
			ep.Scale			= tl.get_scale();
			ep.VBR				= tl.get_VBR();
			ep.VBR_level		= tl.get_VBR_level();
			ep.VBR_max_bitrate	= tl.get_VBR_max_bitrate_kbps();

			return ep;
		}

		/// <summary>
		/// GR[hp[^Zbg
		/// </summary>
		/// <param name="ep">Zbg EncProp \̂ւ̎Q</param>
		/// <returns> true</returns>
		public bool SetEncProps(ref EncProp ep)
		{
			if(tl == null) return false;
			EncProp ep_now = GetEncProps();

			if(ep.Equals(ep_now)) return true;
			EncProp ep_bak = ep_now;

			if(ep_now.ATH_level != ep.ATH_level)
				if(0 != tl.set_ATH_level(ep.ATH_level)) goto error;
			if(ep_now.Ancillary_bits != ep.Ancillary_bits)
				if(0 != tl.set_num_ancillary_bits(ep.Ancillary_bits)) goto error;
			if(ep_now.Bitrate != ep.Bitrate)
				if(0 != tl.set_bitrate(ep.Bitrate)) goto error;
			if(ep_now.Copyright != ep.Copyright)
				if(0 != tl.set_copyright(ep.Copyright)) goto error;
			if(ep_now.CRC != ep.CRC)
				if(0 != tl.set_error_protection(ep.CRC)) goto error;
			if(ep_now.Emphasis != ep.Emphasis)
				if(0 != tl.set_emphasis(ep.Emphasis)) goto error;
			if(ep_now.Energy_level_ext != ep.Energy_level_ext)
				if(0 != tl.set_energy_levels(ep.Energy_level_ext)) goto error;
			if(ep_now.InSamplerate != ep.InSamplerate)
				if(0 != tl.set_in_samplerate(ep.InSamplerate)) goto error;
			if(ep_now.MPEG_mode != ep.MPEG_mode)
				if(0 != tl.set_mode(ep.MPEG_mode)) goto error;
			if(ep_now.MPEG_version != ep.MPEG_version)
				if(0 != tl.set_mpeg_version(ep.MPEG_version)) goto error;
			if(ep_now.Num_channels != ep.Num_channels)
				if(0 != tl.set_num_channels(ep.Num_channels)) goto error;
			if(ep_now.Original != ep.Original)
				if(0 != tl.set_original(ep.Original)) goto error;
			if(ep_now.OutSamplerate != ep.OutSamplerate)
				if(0 != tl.set_out_samplerate(ep.OutSamplerate)) goto error;
			if(ep_now.Padding != ep.Padding)
				if(0 != tl.set_padding(ep.Padding)) goto error;
			if(ep_now.Psymodel != ep.Psymodel)
				if(0 != tl.set_psymodel(ep.Psymodel)) goto error;
			if(ep_now.Quick_count != ep.Quick_count)
				if(0 != tl.set_quick_count(ep.Quick_count)) goto error;
			if(ep_now.Quick_mode != ep.Quick_mode)
				if(0 != tl.set_quick_mode(ep.Quick_mode)) goto error;
			if(ep_now.Scale != ep.Scale)
				if(0 != tl.set_scale(ep.Scale)) goto error;
			if(ep_now.VBR != ep.VBR)
				if(0 != tl.set_VBR(ep.VBR)) goto error;
			if(ep_now.VBR_level != ep.VBR_level)
				if(0 != tl.set_VBR_level(ep.VBR_level)) goto error;
			if(ep_now.VBR_max_bitrate != ep.VBR_max_bitrate)
				if(0 != tl.set_VBR_max_bitrate_kbps(ep.VBR_max_bitrate)) goto error;

			return true;

		error:
			SetEncProps(ref ep_bak);
			return false;
		}
	}

	/// <summary>
	/// R[obN֐
	/// </summary>
	public delegate void ProgressCallback(int inc);
	public delegate void FinishedCallback(object ob);

	/// <summary>
	/// GR[hvpeB[\
	/// </summary>
	public struct EncProp
	{
		public Twolame.MPEG_mode	MPEG_mode;
		public Twolame.MPEG_version	MPEG_version;
		public Twolame.Padding		Padding;
		public Twolame.Emphasis		Emphasis;
		public int	 Psymodel;
		public int	 Num_channels;
		public float Scale;
		public int	 OutSamplerate;
		public int   InSamplerate;
		public int   Bitrate;
		public bool  Energy_level_ext;
		public int   Ancillary_bits;
		public bool  CRC;
		public bool  Copyright;
		public bool  Original;
		public bool  VBR;
		public float VBR_level;
		public float ATH_level;
		public int   VBR_max_bitrate;
		public bool  Quick_mode;
		public int   Quick_count;
	}

}
