using System;
using System.IO;

namespace mp2enc
{
	/// <summary>
	/// WAVt@C[_NX
	/// </summary>
	public class WaveReader
	{
		private short ch; 
		private int sps; 
		private int avrbps; 
		private short balign; 
		private short bps;
		private long  len;
		private FileStream ifs;
		private BinaryReader rdr;

		// p
		private void init()
		{
			ch = balign = bps = 0;
			sps = avrbps = 0;
			len = 0;
		}

		/// <summary>
		/// ʏ̏
		/// </summary>
		public WaveReader()
		{
			init();
		}

		/// <summary>
		/// pXw肵ă[_J
		/// </summary>
		/// <param name="path">t@Cւ̃pX</param>
		public WaveReader(string path)
		{
			Open(path);
		}

		/// <summary>
		/// t@CXg[w肵ă[_J
		/// </summary>
		/// <param name="fs">t@C[_</param>
		public WaveReader(FileStream fs)
		{
			Open(fs);
		}

		/// <summary>
		/// pXw肵ăt@CJ
		/// </summary>
		/// <param name="path">t@Cւ̃pX</param>
		/// <returns> true</returns>
		public bool Open(string path)
		{
			init();

			try
			{
				ifs = new FileStream(path, FileMode.Open, FileAccess.Read);
				if(Open(ifs))
				{
					return true;
				}
				else
				{
					ifs.Close();
					init();
				}
			}
			catch(IOException)
			{
				// faile open failed
				return false;
			}

			return false;
		}

		/// <summary>
		/// t@CXg[w肵
		/// WaveReaderJ
		/// </summary>
		/// <param name="fs">̓t@CXg[</param>
		/// <returns> true</returns>
		public bool Open(FileStream fs)
		{
			try
			{
				rdr = new BinaryReader(fs);
			}
			catch(IOException)
			{
				return false;
			}

			return ParseHeader(rdr);
		}

		/// <summary>
		/// RIFF WAVE t@Cwb_ǂݍ
		/// </summary>
		/// <param name="br">ǂݍރt@C̃oCi[_</param>
		/// <returns> true</returns>
		private bool ParseHeader(BinaryReader br)
		{
			uint sz;

			// "RIFF"
			if(0x46464952 != br.ReadInt32())
				goto Fail;

			sz = br.ReadUInt32();
			// "WAVE"
			if(0x45564157 != br.ReadInt32())
				goto Fail;

			// "fmt "
			if(0x20746d66 != br.ReadInt32())
				goto Fail;

			// byte of fmt chunk
			sz = br.ReadUInt32();
			// format ID
			if(0x0001 != br.ReadInt16())
				goto Fail;
			//ch
			ch = br.ReadInt16();
			//fs
			sps = br.ReadInt32();
			//avr byte per sec
			avrbps = br.ReadInt32();
			//block size
			balign = br.ReadInt16();
			//bit ber sample
			bps = br.ReadInt16();

			sz -= 16;
			byte b;
			for(; sz>0; --sz) b = br.ReadByte();

			// "data"
			if(0x61746164 != br.ReadInt32())
				goto Fail;
			//data len
			len = br.ReadUInt32();

			return true;
		Fail:
			br.Close();
			return false;
		}

		/// <summary>
		/// JĂt@C
		/// </summary>
		public void Close()
		{
			if(rdr != null) rdr.Close();
			if(ifs != null) ifs.Close();
			init();
		}

		/// <summary>
		/// f[^ǂݍ
		/// oCi[_ƓC^[tF[X
		/// </summary>
		/// <param name="buffer">ǂݍރobt@</param>
		/// <param name="index">obt@ł̏݊Jnʒu</param>
		/// <param name="count">ǂݍ݃oCg</param>
		/// <returns>ǂݍ񂾃oCg G[͕̒l</returns>
		public int Read(byte[] buffer, int index, int count)
		{
			if(rdr == null)
				return -1;

			return rdr.Read(buffer, index, count);
		}

		/// <summary>
		/// `l
		/// </summary>
		public short Channel
		{
			get {
				return ch;
			}
		}

		/// <summary>
		/// TvO[g
		/// </summary>
		public int SamplesPerSec
		{
			get {
				return sps;
			}
		}

		/// <summary>
		/// b̕σoCg
		/// </summary>
		public int AvgBytesPerSec
		{
			get {
				return avrbps;
			}
		}

		/// <summary>
		/// ubNTCY
		/// </summary>
		public short BlockAlign
		{
			get 
			{
				return balign;
			}
		}

		/// <summary>
		/// Tv̉𑜓xibitPʁj
		/// </summary>
		public short BitsPerSample
		{
			get 
			{
				return bps;
			}
		}

		/// <summary>
		/// Wave f[^̒ioCgPʁj
		/// </summary>
		public long Length
		{
			get
			{
				return len;
			}
		}
	}
}
